var fs = require('fs');
var q = require('q');
var util = require('util');
var path = require('path');
var cp = require('child_process');
var futil = require('flapjack-util');
var events = require('events');

function SerialQueue() {
  this.fns = [];
  this.executing = false;
}

util.inherits(SerialQueue, events.EventEmitter);

SerialQueue.prototype.push = function(fn) {
  if(this.fns.push(fn) === 1) {
    this.next();
  }
  return this;
};

SerialQueue.prototype.next = function(force) {
  var f = this.fns.shift();
  if(!this.executing || force) {
    if(f) {
      this.executing = true;
      f().then(this.next.bind(this, true), this.stop.bind(this));
    } else {
      this.executing = false;
      this.emit(SerialQueue.Event.DONE);
    }
  }
};

SerialQueue.prototype.stop = function() {
  this.fns = [];
  this.executing = false;
  this.emit.apply(this, [ SerialQueue.Event.DONE ].concat(Array.prototype.slice.call(arguments)));
};

SerialQueue.Event = {
  DONE: 'done'
};

function GulpProject(src) {
  this.src = path.resolve(src);
  if(!this.src || !fs.existsSync(this.src) || !fs.statSync(this.src).isDirectory()) {
    throw util.format('Invalid directory: "%s"', src);
  }
  this.queue = new SerialQueue();

  this.queue.on(SerialQueue.Event.DONE, function() {
    if(this.error) {
      this.emit(GulpProject.Event.BUILD_FAILED, this.error);
    } else {
      this.emit(GulpProject.Event.BUILD_SUCCESSFUL);
    }
  }.bind(this));
}

util.inherits(GulpProject, events.EventEmitter);

GulpProject.prototype.status = function() {
  var status = GulpProject.Status.LAST_BUILD_SUCCESSFUL;
  if(this.queue.executing) {
    status = GulpProject.Status.BUILDING;
  } else if(this.error) {
    status = GulpProject.Status.LAST_BUILD_FAILED;
  }
  return status;
};

GulpProject.prototype.build = function() {
  this.queue.push(function() {
    var built = q.defer();
    var p = cp.fork(path.resolve(__dirname, 'gulp-runner.js'), { cwd: this.src });

    p.on('message', function(m) {
      var message = new futil.GulpMessage(m);
      if(message.data === 'complete' && !this.error) {
        built.resolve();
      } else if(message.type === futil.GulpMessage.Type.ERROR) {
        this.error = message.data;
        built.reject();
      }
    }.bind(this));

    p.send(new futil.GulpMessage({ type: futil.GulpMessage.Type.MESSAGE,
        data: 'start' }));

    return built.promise;
  }.bind(this));

  return this;
};

GulpProject.Event = {
  BUILD_SUCCESSFUL: 'build-successful',
  BUILD_FAILED: 'build-failed'
};
Object.freeze(GulpProject.Event);

GulpProject.Status = {
  BUILDING: 'building',
  LAST_BUILD_SUCCESSFUL: 'last-build-successful',
  LAST_BUILD_FAILED: 'last-build-failed'
};
Object.freeze(GulpProject.Status);

module.exports = GulpProject;
